/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * CayleyManager.hpp
 *
 *  Created on: Sep 8, 2011
 *      Author: dstolee
 */

#ifndef CAYLEYMANAGER_HPP_
#define CAYLEYMANAGER_HPP_

#include "graph.h"
#include "SearchManager.hpp"

class CayleyManager: public SearchManager
{
protected:
	typedef SearchManager super;

	int maxGens;
	int maxN;

	graph_t* g;

	/* DATA */
	int numGens;
	int curGenIndex;
	int curMaxGenerator;
	int* generators;
	int N;
	int R;
	int* point_blocks;
	int* clique;
	int num_available;

	int largestClique();

	int largestClique(int min_v, int cur_size, int cur_max);

	int numberCliques(int x, int y, int to_size);
	int numberCliques(int x, int y, int to_size, int cur_size, int cur_num, int min_v);

	void init();
	int clean();

public:
	/**
	 * Constructor
	 */
	CayleyManager(int maxGens, int maxN);

	/**
	 * Destructor
	 */
	virtual ~CayleyManager();

	/**
	 * pushNext -- deepen the search to the next child
	 * 	of the current node.
	 *
	 * @return the label for the new node. -1 if none.
	 */
	virtual LONG_T pushNext();

	/**
	 * prune -- Perform a check to see if this nod
	 * pushTo -- deepen the search to the specified child
	 * 	of the current node.
	 *
	 * @param child the specified label for the new node
	 * @return the label for the new node. -1 if none, or failed.
	 */
	virtual LONG_T pushTo(LONG_T child);

	/**
	 * pop -- remove the current node and move up the tree.
	 *
	 * @return the label of the node after the pop.
	 * 		This return value is used for validation purposes
	 * 		to check proper implementation of push*() and pop().
	 */
	virtual LONG_T pop();

	/**
	 * prune -- Perform a check to see if this node should be pruned.
	 *
	 * @return 0 if no prune should occur, 1 if prune should occur.
	 */
	virtual int prune();

	/**
	 * isSolution -- Perform a check to see if a solution exists
	 * 		at this point.
	 *
	 * @return 0 if no solution is found, 1 if a solution is found.
	 */
	virtual int isSolution();

	/**
	 * writeSolution -- create a buffer that contains a
	 * 	description of the solution.
	 *
	 * Solution strings start with S.
	 *
	 * @return a string of data allocated with malloc().
	 * 	It will be deleted using free().
	 */
	virtual char* writeSolution();

	/**
	 * writeStatistics -- create a buffer that contains a
	 * 	description of the solution.
	 *
	 * Statistics take the following format in each line:
	 *
	 * T [TYPE] [ID] [VALUE]
	 *
	 * @return a string of data allocated with malloc().
	 * 	It will be deleted using free().
	 */
	virtual char* writeStatistics();

	void reset();
};

#endif /* CAYLEYMANAGER_HPP_ */
