/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * CliquerData.cpp
 *
 *  Created on: Sep 11, 2011
 *      Author: derrickstolee
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "nausparse.h"
#include "gtools.h"
#include "connectivity.hpp"
#include "translation.hpp"
#include "TreeSet.hpp"
#include "cliquer.h"
#include "graph.h"
#include "CliquerData.hpp"

/**
 * cliquer_graph
 */
//	graph_t* cliquer_graph;
void CliquerData::initCliquerGraph()
{
	this->cliquer_graph = graph_new(this->getN());
}

void CliquerData::cleanCliquerGraph()
{
	graph_free(this->cliquer_graph);
	this->cliquer_graph = 0;
}

void CliquerData::addEdgeToCliquerGraph(int i, int j)
{
	GRAPH_ADD_EDGE(this->cliquer_graph, i, j);
	this->cliquer_graph_stack.push(indexOf(i, j));
}

//	std::stack<int> cliquer_graph_stack;
//	std::stack<int> cliquer_graph_sizes;
void CliquerData::snapshotCliquerGraph()
{
	this->cliquer_graph_sizes.push(this->cliquer_graph_stack.size());
}

void CliquerData::rollbackCliquerGraph()
{
	int to_size = this->cliquer_graph_sizes.top();
	this->cliquer_graph_sizes.pop();

	while ( this->cliquer_graph_stack.size() > to_size )
	{
		int index = this->cliquer_graph_stack.top();
		this->cliquer_graph_stack.pop();

		int i, j;
		indexToPair(index, i, j);

		GRAPH_DEL_EDGE(this->cliquer_graph, i, j);
	}
}

void CliquerData::init()
{
	super::init();

	this->initCliquerGraph();
}

void CliquerData::clean()
{
	super::clean();

	this->cleanCliquerGraph();
}

/**
 * Constructor
 */
CliquerData::CliquerData(int r, int n) :
	GraphData(r, n)
{
	this->initCliquerGraph();
}

/**
 * Destructor
 */
CliquerData::~CliquerData() 
{
	this->cleanCliquerGraph();
}

/**
 * snapshot()
 *
 * Freeze the current state in a stack.
 */
void CliquerData::snapshot()
{
	super::snapshot();

	this->snapshotCliquerGraph();
}

/**
 * rollback()
 *
 * Pop the stack and revert the structure to the previous state.
 */
void CliquerData::rollback()
{
	super::rollback();

	this->rollbackCliquerGraph();
}

/**
 * reset()
 *
 * Set the object back to the initial configuration.
 */
void CliquerData::reset()
{
	this->clean();
	this->init();
}

/**
 * augmentEdge()
 *
 * Augment the graph using the following edge change.
 *
 * return FALSE if the constraints are violated.
 */
bool CliquerData::augmentEdge(int i, int j, char type)
{
	if ( type != 0 && type != 1 )
	{
		return false;
	}

	int edge = indexOf(i, j);

	if ( this->getAdjacency(edge) != 2 )
	{
		if ( this->getAdjacency(edge) == type )
		{
			return true;
		}
		else
		{
			this->inconsistent = true;
			return false;
		}
	}

	this->addConnectedVert(i);
	this->addConnectedVert(j);

	bool success = true;

	/** Graph Stuff! **/
	this->addAdjacency(i, j, type);
	this->addEdgeToLayerGraph(i, j, type);

	if ( type == 1 )
	{
		this->addEdgeToOneGraph(i, j);
		this->addEdgeToCliquerGraph(i, j);
	}
	else if ( type == 0 )
	{
		this->addEdgeToZeroGraph(i, j);
	}

	this->addDegree(i, type);
	this->addDegree(j, type);

	/** Sets! **/
	this->removeTwoEdge(edge);

	if ( type == 0 )
	{
		this->addZeroEdge(edge);
	}

	return success;
}

/**
 * getCliquerGraph()
 */
graph_t* CliquerData::getCliquerGraph()
{
	return this->cliquer_graph;
}

/**
 * isConsistent
 */
bool CliquerData::isConsistent()
{
	if ( this->inconsistent )
	{
		return false;
	}

	int max_clique = clique_unweighted_max_weight(this->cliquer_graph, NULL);

	if ( max_clique >= this->r )
	{
		this->inconsistent = true;
		return false;
	}

	if ( this->two_edges->size() > 0 )
	{
		bool added_edges = false;

		/* propagate and/or return false */
		for ( this->two_edges->resetIterator(); this->two_edges->hasNext(); )
		{
			int index = this->two_edges->next();

			int i, j;
			indexToPair(index, i, j);

			GRAPH_ADD_EDGE(this->cliquer_graph, i, j);
			int num_cliques = clique_find_all(this->cliquer_graph, this->r, this->r, FALSE, NULL);
			GRAPH_DEL_EDGE(this->cliquer_graph, i, j);

			if ( num_cliques > 1 )
			{
				return false;
			}
			else if ( num_cliques == 1 )
			{
				/* make this a 0-edge! */
				this->augmentEdge(i, j, 0);

				if ( this->inconsistent )
				{
					return false;
				}

				added_edges = true;
			}
		}

		if ( added_edges )
		{
			int max_clique = clique_unweighted_max_weight(this->cliquer_graph, NULL);

			if ( max_clique >= this->r )
			{
				this->inconsistent = true;
				return false;
			}
		}
	}

	/* propagate and/or return false */
	for ( this->zero_edges->resetIterator(); this->zero_edges->hasNext(); )
	{
		int index = this->zero_edges->next();

		int i, j;
		indexToPair(index, i, j);

		GRAPH_ADD_EDGE(this->cliquer_graph, i, j);
		int num_cliques = clique_find_all(this->cliquer_graph, this->r, this->r, FALSE, NULL);
		GRAPH_DEL_EDGE(this->cliquer_graph, i, j);

		if ( num_cliques != 1 )
		{
			this->inconsistent = true;
			return false;
		}
	}

	return true;
}

/**
 * isUniquelyKrSaturated
 */
bool CliquerData::isUniquelyKrSaturated()
{
	if ( this->inconsistent )
	{
		return false;
	}

	//	int max_clique = clique_unweighted_max_weight(this->cliquer_graph, NULL);
	//
	//	if ( max_clique >= this->r )
	//	{
	//		this->inconsistent = true;
	//		return false;
	//	}

	if ( this->two_edges->size() > 0 )
	{
		return false;
	}

	/* propagate and/or return false */
	//	for ( this->zero_edges->resetIterator(); this->zero_edges->hasNext(); )
	//	{
	//		int index = this->zero_edges->next();
	//
	//		int i, j;
	//		indexToPair(index, i, j);
	//
	//		GRAPH_ADD_EDGE(this->cliquer_graph, i, j);
	//		int num_cliques = clique_find_all(this->cliquer_graph, this->r, this->r, FALSE, NULL);
	//		GRAPH_DEL_EDGE(this->cliquer_graph, i, j);
	//
	//		if ( num_cliques != 1 )
	//		{
	//			return false;
	//		}
	//	}

	return true;
}
