/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * GraphData.cpp
 *
 *  Created on: Sep 11, 2011
 *      Author: derrickstolee
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "nausparse.h"
#include "gtools.h"
#include "connectivity.hpp"
#include "translation.hpp"
#include "TreeSet.hpp"
#include "GraphData.hpp"
#include "SaturationDataSymmetry.hpp"

#define CHECKANDFREE(STUFF) if ( STUFF != 0 ){ free(STUFF); STUFF = 0; }
#define CHECKANDDELETE(STUFF) if ( STUFF != 0 ){ delete STUFF; STUFF = 0; }
#define POPALL(STACK) while (STACK.size() > 0) { STACK.pop(); }

#define SD_DEBUG_MODE
#define SD_DEBUG_PRIORITY 3

#ifdef SD_DEBUG_MODE
#define SD_DEBUG1(PRIORITY,STRING,ARG) if ( PRIORITY <= SD_DEBUG_PRIORITY) printf(STRING,ARG)
#define SD_DEBUG2(PRIORITY,STRING,ARG1,ARG2) if ( PRIORITY <= SD_DEBUG_PRIORITY) printf(STRING,ARG1,ARG2)
#define SD_DEBUG3(PRIORITY,STRING,ARG1,ARG2,ARG3) if ( PRIORITY <= SD_DEBUG_PRIORITY) printf(STRING,ARG1,ARG2,ARG3)
#else
#define SD_DEBUG1(PRIORITY,STRING,ARG)
#define SD_DEBUG2(PRIORITY,STRING,ARG1,ARG2)
#define SD_DEBUG3(PRIORITY,STRING,ARG1,ARG2,ARG3)
#endif

/**
 * layer_graph is the two-layered graph which places edges in each layer
 * 	for type 0 and type 1 edges.
 * It is built to grow as edges and vertices are added.
 * The layers alternate: normal vertex 0 maps to 0 and 1, vert 1 is 2 and 3, etc.
 *
 * The graph is modified by the following operations:
 *  addEdgeToLayerGraph( i,  j,  type)
 *  removeEdgeFromLayerGraph(i, j, type)
 *  addVertexToLayerGraph()
 *  removeVertexFromLayerGraph()
 */
//	sparsegraph* layer_graph;
void GraphData::initLayerGraph()
{
	this->layer_graph = (sparsegraph*) malloc(sizeof(sparsegraph));

	this->layer_graph->nv = 2 * this->n;
	this->layer_graph->nde = 2 * this->n;

	this->layer_graph->vlen = 2 * this->n;
	this->layer_graph->v = (int*) malloc(2 * this->n * sizeof(int));

	this->layer_graph->dlen = 2 * this->n;
	this->layer_graph->d = (int*) malloc(2 * this->n * sizeof(int));

	this->layer_graph->elen = 2 * this->n * this->n;
	this->layer_graph->e = (int*) malloc(2 * this->n * this->n * sizeof(int));
	for ( int i = 0; i < this->layer_graph->elen; i++ )
	{
		this->layer_graph->e[i] = 0;
	}

	for ( int i = 0; i < this->n; i++ )
	{
		this->layer_graph->v[i] = this->n * i;
		this->layer_graph->v[this->n + i] = this->n * this->n + this->n * i;

		this->layer_graph->d[i] = 1;
		this->layer_graph->d[this->n + i] = 1;

		/* put the matching down */
		this->layer_graph->e[this->layer_graph->v[i]] = this->n + i;
		this->layer_graph->e[this->layer_graph->v[this->n + i]] = i;
	}

	this->layer_graph->w = 0;
	this->layer_graph->wlen = 0;
}

void GraphData::cleanLayerGraph()
{
	if ( this->layer_graph != 0 )
	{
		CHECKANDFREE(this->layer_graph->v);
		CHECKANDFREE(this->layer_graph->d);
		CHECKANDFREE(this->layer_graph->e);

		this->layer_graph->vlen = 0;
		this->layer_graph->dlen = 0;
		this->layer_graph->elen = 0;
		this->layer_graph->nv = 0;
		this->layer_graph->nde = 0;

		free(this->layer_graph);
		this->layer_graph = 0;
	}

	POPALL(this->layer_graph_stack);
	POPALL(this->layer_graph_sizes);
}

void GraphData::addEdgeToLayerGraph(int i, int j, char type)
{
	if ( type == 0 )
	{
		int index = indexOf(i, j);

		this->layer_graph->e[this->layer_graph->v[i] + this->layer_graph->d[i]] = j;
		this->layer_graph->e[this->layer_graph->v[j] + this->layer_graph->d[j]] = i;

		this->layer_graph->d[i] = this->layer_graph->d[i] + 1;
		this->layer_graph->d[j] = this->layer_graph->d[j] + 1;

		this->layer_graph->nde = this->layer_graph->nde + 2;

		this->layer_graph_stack.push(index);
	}
	else if ( type == 1 )
	{
		int ni = this->n + i;
		int nj = this->n + j;

		int index = indexOf(ni, nj);

		this->layer_graph->e[this->layer_graph->v[ni] + this->layer_graph->d[ni]] = nj;
		this->layer_graph->e[this->layer_graph->v[nj] + this->layer_graph->d[nj]] = ni;

		this->layer_graph->d[ni] = this->layer_graph->d[ni] + 1;
		this->layer_graph->d[nj] = this->layer_graph->d[nj] + 1;

		this->layer_graph->nde = this->layer_graph->nde + 2;

		this->layer_graph_stack.push(index);
	}
}

//	std::stack<int> layer_graph_stack;
//	std::stack<int> layer_graph_sizes;
void GraphData::snapshotLayerGraph()
{
	this->layer_graph_sizes.push(this->layer_graph_stack.size());
}

void GraphData::rollbackLayerGraph()
{
	int to_size = this->layer_graph_sizes.top();
	this->layer_graph_sizes.pop();

	while ( this->layer_graph_stack.size() > to_size )
	{
		int index = this->layer_graph_stack.top();
		this->layer_graph_stack.pop();

		int vi, vj;

		indexToPair(index, vi, vj);

		this->layer_graph->d[vi] = this->layer_graph->d[vi] - 1;
		this->layer_graph->d[vj] = this->layer_graph->d[vj] - 1;

		this->layer_graph->e[this->layer_graph->v[vi] + this->layer_graph->d[vi]] = 0;
		this->layer_graph->e[this->layer_graph->v[vj] + this->layer_graph->d[vj]] = 0;

		this->layer_graph->nde = this->layer_graph->nde - 2;
	}
}

/**
 * one_graph is the standard graph of just the 1-type edges.
 *
 * The graph is built to grow with vertex and edge additions.
 *
 * The graph is modified by the methods
 * 	addEdgeToOneGraph(i,j)
 *  removeEdgeFromOneGraph(i,j)
 */
//	sparsegraph* one_graph;
void GraphData::initOneGraph()
{
	this->one_graph = (sparsegraph*) malloc(sizeof(sparsegraph));

	SG_INIT(*(this->one_graph));

	this->one_graph->nv = this->n;
	this->one_graph->nde = 0;

	this->one_graph->vlen = this->n;
	this->one_graph->v = (int*) malloc(this->n * sizeof(int));

	this->one_graph->dlen = this->n;
	this->one_graph->d = (int*) malloc(this->n * sizeof(int));

	this->one_graph->elen = this->n * this->n;
	this->one_graph->e = (int*) malloc(this->n * this->n * sizeof(int));
	bzero(this->one_graph->e, this->n * this->n * sizeof(int));

	for ( int i = 0; i < this->n; i++ )
	{
		this->one_graph->v[i] = this->n * i;

		this->one_graph->d[i] = 0;
	}

	this->one_graph->w = 0;
	this->one_graph->wlen = 0;
}

void GraphData::cleanOneGraph()
{
	if ( this->one_graph != 0 )
	{
		CHECKANDFREE(this->one_graph->v);
		CHECKANDFREE(this->one_graph->d);
		CHECKANDFREE(this->one_graph->e);

		this->one_graph->nv = 0;
		this->one_graph->nde = 0;
		this->one_graph->dlen = 0;
		this->one_graph->elen = 0;
		this->one_graph->vlen = 0;

		free(this->one_graph);
		this->one_graph = 0;
	}

	POPALL(this->one_graph_stack);
	POPALL(this->one_graph_sizes);
}

void GraphData::addEdgeToOneGraph(int i, int j)
{
	this->one_graph->e[this->one_graph->v[i] + this->one_graph->d[i]] = j;
	this->one_graph->e[this->one_graph->v[j] + this->one_graph->d[j]] = i;

	this->one_graph->d[i] = this->one_graph->d[i] + 1;
	this->one_graph->d[j] = this->one_graph->d[j] + 1;

	this->one_graph->nde = this->one_graph->nde + 2;

	this->one_graph_stack.push(indexOf(i, j));
}

//	std::stack<int> one_graph_stack;
//	std::stack<int> one_graph_sizes;
void GraphData::snapshotOneGraph()
{
	this->one_graph_sizes.push(this->one_graph_stack.size());
}

void GraphData::rollbackOneGraph()
{
	int to_size = this->one_graph_sizes.top();
	this->one_graph_sizes.pop();

	while ( this->one_graph_stack.size() > to_size )
	{
		int index = this->one_graph_stack.top();
		this->one_graph_stack.pop();

		int i, j;
		indexToPair(index, i, j);

		this->one_graph->d[i] = this->one_graph->d[i] - 1;
		this->one_graph->d[j] = this->one_graph->d[j] - 1;

		this->one_graph->e[this->one_graph->v[i] + this->one_graph->d[i]] = 0;
		this->one_graph->e[this->one_graph->v[j] + this->one_graph->d[j]] = 0;

		this->one_graph->nde = this->one_graph->nde - 2;
	}
}

/**
 * zero_graph is the graph containing only the 0-type edges.
 *
 * The graph is built to grow with vertex and edge additions.
 *
 * The graph is modified by the methods
 * 	addEdgeToZeroGraph(i,j)
 *  removeEdgeFromZeroGraph(i,j)
 */
//	sparsegraph* zero_graph;
void GraphData::initZeroGraph()
{
	this->zero_graph = (sparsegraph*) malloc(sizeof(sparsegraph));

	SG_INIT(*(this->zero_graph));

	this->zero_graph->nv = this->n;
	this->zero_graph->nde = 0;

	this->zero_graph->vlen = this->n;
	this->zero_graph->v = (int*) malloc(this->n * sizeof(int));

	this->zero_graph->dlen = this->n;
	this->zero_graph->d = (int*) malloc(this->n * sizeof(int));

	this->zero_graph->elen = this->n * this->n;
	this->zero_graph->e = (int*) malloc(this->n * this->n * sizeof(int));
	bzero(this->zero_graph->e, this->n * this->n * sizeof(int));

	for ( int i = 0; i < this->n; i++ )
	{
		this->zero_graph->v[i] = this->n * i;
		this->zero_graph->d[i] = 0;
	}

	this->zero_graph->w = 0;
	this->zero_graph->wlen = 0;
}

void GraphData::cleanZeroGraph()
{
	if ( this->zero_graph != 0 )
	{
		CHECKANDFREE(this->zero_graph->v);
		CHECKANDFREE(this->zero_graph->d);
		CHECKANDFREE(this->zero_graph->e);

		this->zero_graph->nv = 0;
		this->zero_graph->nde = 0;
		this->zero_graph->dlen = 0;
		this->zero_graph->elen = 0;
		this->zero_graph->vlen = 0;

		free(this->zero_graph);
		this->zero_graph = 0;
	}

	POPALL(this->zero_graph_stack);
	POPALL(this->zero_graph_sizes);
}

void GraphData::addEdgeToZeroGraph(int i, int j)
{
	this->zero_graph->e[this->zero_graph->v[i] + this->zero_graph->d[i]] = j;
	this->zero_graph->e[this->zero_graph->v[j] + this->zero_graph->d[j]] = i;

	this->zero_graph->d[i] = this->zero_graph->d[i] + 1;
	this->zero_graph->d[j] = this->zero_graph->d[j] + 1;

	this->zero_graph->nde = this->zero_graph->nde + 2;

	this->zero_graph_stack.push(indexOf(i, j));
}

//	std::stack<int> zero_graph_stack;
//	std::stack<int> zero_graph_sizes;
void GraphData::snapshotZeroGraph()
{
	this->zero_graph_sizes.push(this->zero_graph_stack.size());
}

void GraphData::rollbackZeroGraph()
{
	int to_size = this->zero_graph_sizes.top();
	this->zero_graph_sizes.pop();

	while ( this->zero_graph_stack.size() > to_size )
	{
		int index = this->zero_graph_stack.top();
		this->zero_graph_stack.pop();

		int i, j;
		indexToPair(index, i, j);

		this->zero_graph->d[i] = this->zero_graph->d[i] - 1;
		this->zero_graph->d[j] = this->zero_graph->d[j] - 1;

		this->zero_graph->e[this->zero_graph->v[i] + this->zero_graph->d[i]] = 0;
		this->zero_graph->e[this->zero_graph->v[j] + this->zero_graph->d[j]] = 0;

		this->zero_graph->nde = this->zero_graph->nde - 2;
	}
}

/**
 * The adjacency matrix stores edge types of 0, 1, 2, and 3.
 *
 * 0: A known non-edge. It MUST have a completion.
 * 1: A known edge.
 * 2: An unknown edge.
 * 3: An edge pair which uses a vertex that is not currently in the graph.
 */
//	char* adjmat;
void GraphData::initAdjacencyMatrix()
{
	this->adjmat = (char*) malloc(this->n * this->n);

	for ( int i = 0; i < this->n * this->n; i++ )
	{
		this->adjmat[i] = 2;
	}
}

void GraphData::cleanAdjacencyMatrix()
{
	CHECKANDFREE(this->adjmat);
	POPALL(this->adjacency_stack);
	POPALL(this->adjacency_sizes);
}

bool GraphData::addAdjacency(int i, int j, char type)
{
	int index = indexOf(i, j);

	if ( this->adjmat[index] == 2 )
	{
		this->adjmat[index] = type;
		this->adjacency_stack.push(index);
		return true;
	}

	return false;
}

//	std::stack<int> adjacency_stack;
//	std::stack<int> adjacency_sizes;
void GraphData::snapshotAdjacencies()
{
	this->adjacency_sizes.push(this->adjacency_stack.size());
}

void GraphData::rollbackAdjacencies()
{
	int to_size = this->adjacency_sizes.top();
	this->adjacency_sizes.pop();

	while ( this->adjacency_stack.size() > to_size )
	{
		int index = this->adjacency_stack.top();
		this->adjacency_stack.pop();

		this->adjmat[index] = 2;
	}
}

/**
 * The degree sequence stores two types of degrees: 0 and 1.
 *
 * It is a 2-dimensional array, maxN x 2, packed into a single character list.
 *
 * The data type is char since we will never reach close to even 50 vertices.
 *
 * The getDegree(i,type) and setDegree(i,type) methods modify this sequence.
 */
//	char* degseq;
void GraphData::initDegreeSequence()
{
	this->degseq = (char*) malloc(2 * this->n);
	bzero(this->degseq, 2 * this->n);
}

void GraphData::cleanDegreeSequence()
{
	CHECKANDFREE(this->degseq);
	POPALL(this->degree_stack);
	POPALL(this->degree_sizes);
}

void GraphData::addDegree(int i, int type)
{
	int pos = this->n * type + i;
	this->degseq[pos] = this->degseq[pos] + 1;
	this->degree_stack.push(pos);
}

//	std::stack<int> degree_stack;
//	std::stack<int> degree_sizes;
void GraphData::snapshotDegrees()
{
	this->degree_sizes.push(this->degree_stack.size());
}

void GraphData::rollbackDegrees()
{
	int to_size = this->degree_sizes.top();
	this->degree_sizes.pop();

	while ( this->degree_stack.size() > to_size )
	{
		int pos = this->degree_stack.top();
		this->degree_stack.pop();

		this->degseq[pos] = this->degseq[pos] - 1;
	}
}

/**
 * two_edges
 */
//	Set* two_edges;
//	Set* zero_edges;
void GraphData::initTwoEdges()
{
	this->two_edges = new TreeSet();
	this->zero_edges = new TreeSet();

	for ( int i = 0; i < this->nchoose2; i++ )
	{
		this->two_edges->add(i);
	}
}

void GraphData::cleanTwoEdges()
{
	CHECKANDDELETE(this->two_edges);
	CHECKANDDELETE(this->zero_edges);

	POPALL(this->two_edges_stack);
	POPALL(this->two_edges_sizes);
	POPALL(this->zero_edges_stack);
	POPALL(this->zero_edges_sizes);
}

void GraphData::removeTwoEdge(int index)
{
	if ( this->two_edges->contains(index) == 1 )
	{
		this->two_edges->remove(index);
		this->two_edges_stack.push(index);
	}
}

void GraphData::addZeroEdge(int index)
{
	if ( this->zero_edges->contains(index) == 0 )
	{
		this->zero_edges->add(index);
		this->zero_edges_stack.push(index);
	}
}

//	std::stack<int> two_edges_stack;
//	std::stack<int> two_edges_sizes;
//	std::stack<int> zero_edges_stack;
//	std::stack<int> zero_edges_sizes;
void GraphData::snapshotEdgeSets()
{
	this->two_edges_sizes.push(this->two_edges_stack.size());
	this->zero_edges_sizes.push(this->zero_edges_stack.size());
}

void GraphData::rollbackEdgeSets()
{
	int to_size1 = this->two_edges_sizes.top();
	this->two_edges_sizes.pop();

	while ( this->two_edges_stack.size() > to_size1 )
	{
		int edge = this->two_edges_stack.top();
		this->two_edges_stack.pop();

		this->two_edges->add(edge);
	}

	int to_size2 = this->zero_edges_sizes.top();
	this->zero_edges_sizes.pop();

	while ( this->zero_edges_stack.size() > to_size2 )
	{
		int edge = this->zero_edges_stack.top();
		this->zero_edges_stack.pop();

		this->zero_edges->remove(edge);
	}
}

/**
 * connected_verts
 *
 * This is the list of vertices which have been included in the augmentations.
 */
//	Set* connected_verts;
void GraphData::initConnectedVerts()
{
	this->connected_verts = new TreeSet();
}

void GraphData::cleanConnectedVerts()
{
	CHECKANDDELETE(this->connected_verts);
	POPALL(this->connected_verts_snapshots);
	POPALL(this->connected_verts_stack);
}

void GraphData::addConnectedVert(int v)
{
	if ( this->connected_verts->contains(v) == 0 )
	{
		this->connected_verts->add(v);
		this->connected_verts_stack.push(v);
	}
}

//	std::stack<int> connected_verts_stack;
//	std::stack<int> connected_verts_snapshots;
void GraphData::snapshotConnectedVerts()
{
	this->connected_verts_snapshots.push(this->connected_verts_stack.size());
}

void GraphData::rollbackConnectedVerts()
{
	int to_size = this->connected_verts_snapshots.top();
	this->connected_verts_snapshots.pop();

	while ( this->connected_verts_stack.size() > to_size )
	{
		int v = this->connected_verts_stack.top();
		this->connected_verts_stack.pop();

		this->connected_verts->remove(v);
	}
}

void GraphData::init()
{
	this->inconsistent = false;
	this->initLayerGraph();
	this->initOneGraph();
	this->initZeroGraph();
	this->initAdjacencyMatrix();
	this->initDegreeSequence();
	this->initTwoEdges();
	this->initConnectedVerts();
}

void GraphData::clean()
{
	this->cleanLayerGraph();
	this->cleanOneGraph();
	this->cleanZeroGraph();
	this->cleanAdjacencyMatrix();
	this->cleanDegreeSequence();
	this->cleanTwoEdges();
	this->cleanConnectedVerts();
}

/**
 * Constructor
 */
GraphData::GraphData(int r, int N)
{
	this->r = r;
	this->n = N;

	this->s = r - 2;
	this->nchoose2 = nChooseK(N, 2);
	this->nchooser = nChooseK(N, r);
	this->nchooses = nChooseK(N, r - 2);

	this->init();
}

/**
 * Destructor
 */
GraphData::~GraphData()
{
	this->clean();
}

/**
 * snapshot()
 *
 * Freeze the current state in a stack.
 */
void GraphData::snapshot()
{
	this->snapshotLayerGraph();
	this->snapshotOneGraph();
	this->snapshotZeroGraph();
	this->snapshotAdjacencies();
	this->snapshotDegrees();
	this->snapshotEdgeSets();
	this->snapshotConnectedVerts();
}

/**
 * rollback()
 *
 * Pop the stack and revert the structure to the previous state.
 */
void GraphData::rollback()
{
	this->rollbackLayerGraph();
	this->rollbackOneGraph();
	this->rollbackZeroGraph();
	this->rollbackAdjacencies();
	this->rollbackDegrees();
	this->rollbackEdgeSets();
	this->rollbackConnectedVerts();
	this->inconsistent = false;
}

/**
 * reset()
 *
 * Set the object back to the initial configuration.
 */
void GraphData::reset()
{
	this->clean();
	this->init();
}

/**
 * augmentEdge()
 *
 * Augment the graph using the following edge change.
 *
 * return FALSE if the constraints are violated.
 */
bool GraphData::augmentEdge(int index, char type)
{
	int i, j;
	indexToPair(index, i, j);

	return this->augmentEdge(i, j, type);
}

/**
 * augmentFill(set)
 *
 * Make all of the given edges in 'set' (a -1 terminated array)
 * 	become the given type.
 */
bool GraphData::augmentEdges(int* set, char type)
{
	int i = 0;
	bool result = true;
	while ( result && set[i] >= 0 )
	{
		result = this->augmentEdge(set[i], type);
		i++;
	}

	return result;
}

/**
 * augmentCompletion()
 *
 * Augment by making a 0-edge {i,j} with a completion (an r-2 subset index).
 */
bool GraphData::augmentCompletion(int i, int j, int completion)
{
	//	int index = indexOf(i, j);

	int* set = (int*) malloc((this->r - 2) * sizeof(int));

	indexToSet(this->r - 2, completion, set);

	bool result = true;

	for ( int x = 0; result && x < this->r - 2; x++ )
	{
		result = result && this->augmentEdge(i, set[x], 1);
		result = result && this->augmentEdge(j, set[x], 1);

		for ( int y = x + 1; result && y < this->r - 2; y++ )
		{
			result = result && this->augmentEdge(set[x], set[y], 1);
		}
	}

	result = result && this->augmentEdge(i, j, 0);

	free(set);
	set = 0;

	return result;
}

/**
 * augmentCompletion()
 *
 * Augment by making a 0-edge {i,j} with a completion (an r-2 subset index).
 */
bool GraphData::augmentCompletion(int index, int completion)
{
	int vi, vj;
	indexToPair(index, vi, vj);

	return this->augmentCompletion(vi, vj, completion);
}

/**
 * getLayeredGraph()
 */
sparsegraph* GraphData::getLayeredGraph()
{
	return this->layer_graph;
}

/**
 * getOneGraph()
 */
sparsegraph* GraphData::getOneGraph()
{
	return this->one_graph;
}

/**
 * getZeroGraph()
 */
sparsegraph* GraphData::getZeroGraph()
{
	return this->zero_graph;
}

/**
 * getOneGraphString()
 *
 * NOTE: DO NOT FREE() THIS STRING!
 */
char* GraphData::getOneGraphString()
{
	return sgtos6(this->one_graph);
}

/**
 * getN Return the current number of vertices.
 */
int GraphData::getN()
{
	return this->n;
}

/**
 * getAdjacency return the adjacency for this pair
 * index or vertex pair.
 */
char GraphData::getAdjacency(int index)
{
	return this->adjmat[index];
}

char GraphData::getAdjacency(int i, int j)
{
	return this->adjmat[indexOf(i, j)];
}

/**
 * getNum2Edges
 */
int GraphData::getNum2Edges()
{
	return this->two_edges->size();
}

char GraphData::getDegree(int i, int type)
{
	return this->degseq[type * this->n + i];
}

bool GraphData::isInConnected(int v)
{
	if ( this->connected_verts->contains(v) == 1 )
	{
		return true;
	}

	return false;
}

/**
 * isUniquelyKrSaturated
 */
bool GraphData::isUniquelyKrSaturated()
{
	if ( this->inconsistent )
	{
		return false;
	}

	if ( this->two_edges->size() > 0 )
	{
		return false;
	}

	return true;
}

/**
 * isConsistent
 */
bool GraphData::isConsistent()
{
	return !(this->inconsistent);
}
