/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * GraphData.hpp
 *
 *  Created on: Sep 11, 2011
 *      Author: derrickstolee
 */

#ifndef GRAPHDATA_HPP_
#define GRAPHDATA_HPP_

#include <stack>
#include "nausparse.h"
#include "Set.hpp"

class GraphData
{
protected:

	/**
	 * n the number of vertices.
	 */
	int n;

	/**
	 * r -- Specify which K_r we are using for unique saturation
	 */
	int r;
	int s; /* r-2 */

	int nchoose2;
	int nchooser;
	int nchooses;

	/**
	 * inconsistent -- did the last augmentation create an inconsistency?
	 */
	bool inconsistent;

	/**
	 * layer_graph is the two-layered graph which places edges in each layer
	 * 	for type 0 and type 1 edges.
	 * It is built to grow as edges and vertices are added.
	 * The layers alternate: normal vertex 0 maps to 0 and 1, vert 1 is 2 and 3, etc.
	 *
	 * The graph is modified by the following operations:
	 *  addEdgeToLayerGraph( i,  j,  type)
	 *  removeEdgeFromLayerGraph(i, j, type)
	 *  addVertexToLayerGraph()
	 *  removeVertexFromLayerGraph()
	 */
	sparsegraph* layer_graph;
	void initLayerGraph();
	void cleanLayerGraph();
	void addEdgeToLayerGraph(int i, int j, char type);
	std::stack<int> layer_graph_stack;
	std::stack<int> layer_graph_sizes;
	void snapshotLayerGraph();
	void rollbackLayerGraph();

	/**
	 * one_graph is the standard graph of just the 1-type edges.
	 *
	 * The graph is built to grow with vertex and edge additions.
	 *
	 * The graph is modified by the methods
	 * 	addEdgeToOneGraph(i,j)
	 *  removeEdgeFromOneGraph(i,j)
	 */
	sparsegraph* one_graph;
	void initOneGraph();
	void cleanOneGraph();
	void addEdgeToOneGraph(int i, int j);
	std::stack<int> one_graph_stack;
	std::stack<int> one_graph_sizes;
	void snapshotOneGraph();
	void rollbackOneGraph();

	/**
	 * zero_graph is the graph containing only the 0-type edges.
	 *
	 * The graph is built to grow with vertex and edge additions.
	 *
	 * The graph is modified by the methods
	 * 	addEdgeToZeroGraph(i,j)
	 *  removeEdgeFromZeroGraph(i,j)
	 */
	sparsegraph* zero_graph;
	void initZeroGraph();
	void cleanZeroGraph();
	void addEdgeToZeroGraph(int i, int j);
	std::stack<int> zero_graph_stack;
	std::stack<int> zero_graph_sizes;
	void snapshotZeroGraph();
	void rollbackZeroGraph();

	/**
	 * The adjacency matrix stores edge types of 0, 1, 2, and 3.
	 *
	 * 0: A known non-edge. It MUST have a completion.
	 * 1: A known edge.
	 * 2: An unknown edge.
	 * 3: An edge pair which uses a vertex that is not currently in the graph.
	 */
	char* adjmat;
	void initAdjacencyMatrix();
	void cleanAdjacencyMatrix();
	bool addAdjacency(int i, int j, char type);
	std::stack<int> adjacency_stack;
	std::stack<int> adjacency_sizes;
	void snapshotAdjacencies();
	void rollbackAdjacencies();

	/**
	 * The degree sequence stores two types of degrees: 0 and 1.
	 *
	 * It is a 2-dimensional array, maxN x 2, packed into a single character list.
	 *
	 * The data type is char since we will never reach close to even 50 vertices.
	 *
	 * The getDegree(i,type) and setDegree(i,type) methods modify this sequence.
	 */
	char* degseq;
	void initDegreeSequence();
	void cleanDegreeSequence();
	void addDegree(int i, int type);
	std::stack<int> degree_stack;
	std::stack<int> degree_sizes;
	void snapshotDegrees();
	void rollbackDegrees();


	/**
	 * two_edges
	 */
	Set* two_edges;
	Set* zero_edges;
	void initTwoEdges();
	void cleanTwoEdges();
	void removeTwoEdge(int index);
	void addZeroEdge(int index);
	std::stack<int> two_edges_stack;
	std::stack<int> two_edges_sizes;
	std::stack<int> zero_edges_stack;
	std::stack<int> zero_edges_sizes;
	void snapshotEdgeSets();
	void rollbackEdgeSets();

	/**
	 * connected_verts
	 *
	 * This is the list of vertices which have been included in the augmentations.
	 */
	Set* connected_verts;
	void initConnectedVerts();
	void cleanConnectedVerts();
	void addConnectedVert(int v);
	std::stack<int> connected_verts_stack;
	std::stack<int> connected_verts_snapshots;
	void snapshotConnectedVerts();
	void rollbackConnectedVerts();

	virtual void init();
	virtual void clean();

public:
	/**
	 * Constructor
	 */
	GraphData(int n, int r);

	/**
	 * Destructor
	 */
	~GraphData();

	/**
	 * snapshot()
	 *
	 * Freeze the current state in a stack.
	 */
	virtual void snapshot();

	/**
	 * rollback()
	 *
	 * Pop the stack and revert the structure to the previous state.
	 */
	virtual void rollback();

	/**
	 * reset()
	 *
	 * Set the object back to the initial configuration.
	 */
	virtual void reset();

	/**
	 * augmentEdge()
	 *
	 * Augment the graph using the following edge change.
	 *
	 * return FALSE if the constraints are violated.
	 */
	virtual bool augmentEdge(int i, int j, char type) = 0;

	/**
	 * augmentEdge()
	 *
	 * Augment the graph using the following edge change.
	 *
	 * return FALSE if the constraints are violated.
	 */
	virtual bool augmentEdge(int index, char type);

	/**
	 * augmentFill(set)
	 *
	 * Make all of the given edges in 'set' (a -1 terminated array)
	 * 	become the given type.
	 */
	virtual bool augmentEdges(int* set, char type);

	/**
	 * augmentCompletion()
	 *
	 * Augment by making a 0-edge {i,j} with a completion (an r-2 subset index).
	 */
	virtual bool augmentCompletion(int i, int j, int completion);

	/**
	 * augmentCompletion()
	 *
	 * Augment by making a 0-edge {i,j} with a completion (an r-2 subset index).
	 */
	virtual bool augmentCompletion(int index, int completion);

	/**
	 * getLayeredGraph()
	 */
	virtual sparsegraph* getLayeredGraph();

	/**
	 * getOneGraph()
	 */
	virtual sparsegraph* getOneGraph();

	/**
	 * getZeroGraph()
	 */
	virtual sparsegraph* getZeroGraph();

	/**
	 * getOneGraphString()
	 *
	 * NOTE: DO NOT FREE() THIS STRING!
	 */
	virtual char* getOneGraphString();

	/**
	 * isConsistent
	 */
	virtual bool isConsistent();

	/**
	 * isUniquelyKrSaturated
	 */
	virtual bool isUniquelyKrSaturated();

	/**
	 * getN Return the current number of vertices.
	 */
	virtual int getN();

	/**
	 * getAdjacency return the adjacency for this pair
	 * index or vertex pair.
	 */
	virtual char getAdjacency(int index);
	virtual char getAdjacency(int i, int j);

	/**
	 * getNum2Edges
	 */
	virtual int getNum2Edges();

	virtual char getDegree(int i, int type);

	virtual bool isInConnected(int v);
};

#endif /* GRAPHDATA_HPP_ */
