/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * OrbitalBranchingManager.cpp
 *
 *  Created on: Aug 24, 2011
 *      Author: dstolee
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "SaturationData.hpp"
#include "SearchManager.hpp"
#include "CliquerData.hpp"
#include "SaturationNode.hpp"
#include "SaturationBranchingRule.hpp"
#include "OrbitalBranchingManager.hpp"
#include "SaturationDataSymmetry.hpp"
#include "translation.hpp"

/**
 * initializeNode
 *
 * Run the symmetry methods on this node
 */
void OrbitalBranchingManager::initializeNode(SaturationNode* node, int& orbit)
{
	if ( node->numOrbits <= 0 )
	{
		/* orbit reps are calculated within the SaturationData object */
		/* UGLY :( */
		computeUnassignedOrbits(this->satdata, node);
	}

	if ( node->numOrbits == 0 )
	{
		orbit = -1;
		return;
	}

	orbit = this->rule->selectOrbitIndex(this->satdata, node);

	node->curOrbit = orbit;

	computeStabilizedOrbits(this->r, this->satdata, node->orbitReps[orbit], node);
}

/**
 * getInitialOrbitCompleter
 *
 * Pick the orbit and completer to START the augmentations.
 */
void OrbitalBranchingManager::getInitialOrbitCompleter(SaturationNode* node, int& orbit, int& completer)
{
	completer = 0;
}

/**
 * getNextOrbitCompleter
 *
 * Given the current node, orbit, and completer,
 * 	use the mode to decide the next orbit and completer.
 */
void OrbitalBranchingManager::getNextOrbitCompleter(SaturationNode* node, int& orbit, int& completer)
{
	completer = completer + 1;

	if ( completer > node->curNumCompleters )
	{
		completer = -1;
	}
}

/**
 * Constructor
 */
OrbitalBranchingManager::OrbitalBranchingManager(int r, int maxN, SaturationBranchingRule* rule, bool use_cliquer)
{
	this->r = r;
	this->maxN = maxN;
	this->augmentation_succeeded = false;
	this->rule = rule;

	if ( use_cliquer )
	{
		this->satdata = new CliquerData(r, maxN);
	}
	else
	{
		this->satdata = new SaturationData(r, maxN);
	}

	this->satdata->snapshot();

	/* first augmentation a trivial matter! {r-2,r-1} -completesto-> {0,1,2,...,r-3} */
	this->satdata->augmentCompletion(r - 2, r - 1, 0);

	this->satdata->snapshot();

	if ( this->root != 0 )
	{
		delete this->root;
	}

	this->root = new SaturationNode(0);

}

/**
 * Destructor
 */
OrbitalBranchingManager::~OrbitalBranchingManager()
{
	this->rule = 0;
	delete this->satdata;
	this->satdata = 0;
}

/**
 * pushNext -- deepen the search to the next child
 * 	of the current node.
 *
 * @return the label for the new node. -1 if none.
 */
LONG_T OrbitalBranchingManager::pushNext()
{
	/* get 'top' node */
	SaturationNode* parent = 0;
	if ( this->stack.size() > 0 )
	{
		parent = (SaturationNode*) this->stack.back();
	}
	else
	{
		parent = (SaturationNode*) this->root;
	}

	int orbit = 0;
	int completer = 0;

	if ( parent->numOrbits < 0 )
	{
		this->initializeNode(parent, orbit);
		this->getInitialOrbitCompleter(parent, orbit, completer);

		if ( orbit < 0 )
		{
			return -1;
		}

		parent->curCompleterIndex = completer;
	}
	else
	{
		completer = parent->curCompleterIndex;
		this->getNextOrbitCompleter(parent, orbit, completer);
		parent->curCompleterIndex = completer;
	}

	if ( completer < 0 || completer > parent->curNumCompleters )
	{
		return -1;
	}

	LONG_T child = completer;

	return this->pushTo(child);
}

/**
 * prune -- Perform a check to see if this nod
 * pushTo -- deepen the search to the specified child
 * 	of the current node.
 *
 * @param child the specified label for the new node
 * @return the label for the new node. -1 if none, or failed.
 */
LONG_T OrbitalBranchingManager::pushTo(LONG_T child)
{
	/* make sure we will need a prune */
	if ( child < 0 )
	{
		return -1;
	}

	SaturationNode* parent = 0;
	if ( this->stack.size() > 0 )
	{
		parent = (SaturationNode*) this->stack.back();
	}
	else
	{
		parent = (SaturationNode*) this->root;
	}

	if ( parent->numOrbits <= 0 )
	{
		this->initializeNode(parent, parent->curOrbit);
		//		this->getInitialOrbitCompleter(parent, parent->curOrbit, parent->curCompleterIndex);

		if ( parent->curOrbit < 0 )
		{
			return -1;
		}
	}

	parent->curCompleterIndex = (int) child;
	parent->curChild = child;

	int orbit = parent->curOrbit;
	int completer = (int) child;

	SaturationNode* node = new SaturationNode(child);
	this->stack.push_back(node);

	int vi, vj;
	indexToPair(parent->orbitReps[orbit], vi, vj);

	this->augmentation_succeeded = true;
	if ( completer < parent->curNumCompleters )
	{
		this->satdata->snapshot();
		//		printf(">> augmenting with %d %d to { ", vi, vj);

		//		int* set = (int*) malloc(this->r * sizeof(int));
		//		indexToSet(this->r - 2, parent->curOrbitCompleterReps[completer], set);
		//		for ( int i = 0; i < this->r - 2; i++ )
		//		{
		//			printf("%d ", set[i]);
		//		}
		//
		//		printf("}\n");

		this->augmentation_succeeded = this->satdata->augmentCompletion(vi, vj,
				parent->curOrbitCompleterReps[completer]);
		//		free(set);
	}
	else if ( completer == parent->curNumCompleters )
	{
		this->satdata->snapshot();
		this->augmentation_succeeded = this->satdata->augmentEdges(parent->orbitSets[orbit], 1);
	}
	else
	{
		return -1;
	}

	return child;
}

/**
 * pop -- remove the current node and move up the tree.
 *
 * @return the label of the node after the pop.
 * 		This return value is used for validation purposes
 * 		to check proper implementation of push*() and pop().
 */
LONG_T OrbitalBranchingManager::pop()
{
	this->satdata->rollback();

	if ( this->stack.size() > 0 )
	{
		SearchNode* node = this->stack.back();
		this->stack.pop_back();

		LONG_T label = node->label;
		delete node;
		node = 0;

		return label;
	}

	return 0;
}

/**
 * prune -- Perform a check to see if this node should be pruned.
 *
 * @return 0 if no prune should occur, 1 if prune should occur.
 */
int OrbitalBranchingManager::prune()
{
	if ( this->augmentation_succeeded == false )
	{
		/* totally dependent on augmentation! */
		this->augmentation_succeeded = true;
		return 1;
	}

	if ( this->satdata->isConsistent() == false )
	{
		return 1;
	}

	return 0;
}

/**
 * isSolution -- Perform a check to see if a solution exists
 * 		at this point.
 *
 * @return 0 if no solution is found, 1 if a solution is found.
 */
int OrbitalBranchingManager::isSolution()
{
	if ( this->augmentation_succeeded && this->satdata->getNum2Edges() == 0 && this->satdata->isUniquelyKrSaturated() )
	{
		return 1;
	}

	return 0;
}

/**
 * writeSolution -- create a buffer that contains a
 * 	description of the solution.
 *
 * Solution strings start with S.
 *
 * @return a string of data allocated with malloc().
 * 	It will be deleted using free().
 */
char* OrbitalBranchingManager::writeSolution()
{
	printf("%s\n", this->satdata->getOneGraphString());

	return 0;
}

/**
 * writeStatistics -- create a buffer that contains a
 * 	description of the solution.
 *
 * Statistics take the following format in each line:
 *
 * T [TYPE] [ID] [VALUE]
 *
 * @return a string of data allocated with malloc().
 * 	It will be deleted using free().
 */
char* OrbitalBranchingManager::writeStatistics()
{
	return super::writeBaseStats();
}

void OrbitalBranchingManager::reset()
{
	this->satdata->reset();

	this->satdata->snapshot();

	/* first augmentation a trivial matter! {r-2,r-1} -completesto-> {0,1,2,...,r-3} */
	this->satdata->augmentCompletion(r - 2, r - 1, 0);

	this->satdata->snapshot();

	if ( this->root != 0 )
	{
		delete this->root;
	}

	this->root = new SaturationNode(0);
}

/**
 * Initialize to a custom starting point.
 */
void OrbitalBranchingManager::initToGR6N20A()
{
	this->satdata->reset();

	this->satdata->augmentEdge(0, 1, 1);

	for ( int i = 0; i < 9; i++ )
	{
		this->satdata->augmentEdge(0, 2 + i);
		this->satdata->augmentEdge(1, 9 + i);

		this->satdata->augmentEdge(2 + i, 2 + ((i + 1) % 9), 1);
		this->satdata->augmentEdge(2 + i, 2 + ((i + 2) % 9), 1);
		this->satdata->augmentEdge(2 + i, 2 + ((i + 3) % 9), 1);

		this->satdata->augmentEdge(9 + i, 9 + ((i + 1) % 9), 1);
		this->satdata->augmentEdge(9 + i, 9 + ((i + 2) % 9), 1);
		this->satdata->augmentEdge(9 + i, 9 + ((i + 3) % 9), 1);
	}

	this->satdata->snapshot();
}
