/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * OrbitalBranchingManager.hpp
 *
 *  Created on: Aug 24, 2011
 *      Author: dstolee
 */

#ifndef ORBITALBRANCHINGMANAGER_HPP_
#define ORBITALBRANCHINGMANAGER_HPP_

#include "SaturationData.hpp"
#include "SearchManager.hpp"
#include "SaturationNode.hpp"
#include "SaturationBranchingRule.hpp"
#include "GraphData.hpp"

class OrbitalBranchingManager: public SearchManager
{

protected:
	typedef SearchManager super;

	SaturationBranchingRule* rule;

	int r;
	int maxN;

	/**
	 * All data is stored in a SaturationData object
	 */
	GraphData* satdata;

	/**
	 * augmentation_succeeded
	 */
	bool augmentation_succeeded;

	/**
	 * initializeNode
	 *
	 * Run the symmetry methods on this node and select an orbit
	 */
	void initializeNode(SaturationNode* node, int& orbit);

	/**
	 * getInitialOrbitCompleter
	 *
	 * Pick the orbit and completer to START the augmentations.
	 */
	void getInitialOrbitCompleter(SaturationNode* node, int& orbit, int& completer);

	/**
	 * getNextOrbitCompleter
	 *
	 * Given the current node, orbit, and completer,
	 * 	use the mode to decide the next orbit and completer.
	 */
	void getNextOrbitCompleter(SaturationNode* node, int& orbit, int& completer);

public:
	/**
	 * Constructor
	 */
	OrbitalBranchingManager(int r, int maxN, SaturationBranchingRule* rule, bool use_cliquer);

	/**
	 * Destructor
	 */
	virtual ~OrbitalBranchingManager();

	/**
	 * pushNext -- deepen the search to the next child
	 * 	of the current node.
	 *
	 * @return the label for the new node. -1 if none.
	 */
	virtual LONG_T pushNext();

	/**
	 * prune -- Perform a check to see if this nod
	 * pushTo -- deepen the search to the specified child
	 * 	of the current node.
	 *
	 * @param child the specified label for the new node
	 * @return the label for the new node. -1 if none, or failed.
	 */
	virtual LONG_T pushTo(LONG_T child);

	/**
	 * pop -- remove the current node and move up the tree.
	 *
	 * @return the label of the node after the pop.
	 * 		This return value is used for validation purposes
	 * 		to check proper implementation of push*() and pop().
	 */
	virtual LONG_T pop();

	/**
	 * prune -- Perform a check to see if this node should be pruned.
	 *
	 * @return 0 if no prune should occur, 1 if prune should occur.
	 */
	virtual int prune();

	/**
	 * isSolution -- Perform a check to see if a solution exists
	 * 		at this point.
	 *
	 * @return 0 if no solution is found, 1 if a solution is found.
	 */
	virtual int isSolution();

	/**
	 * writeSolution -- create a buffer that contains a
	 * 	description of the solution.
	 *
	 * Solution strings start with S.
	 *
	 * @return a string of data allocated with malloc().
	 * 	It will be deleted using free().
	 */
	virtual char* writeSolution();

	/**
	 * writeStatistics -- create a buffer that contains a
	 * 	description of the solution.
	 *
	 * Statistics take the following format in each line:
	 *
	 * T [TYPE] [ID] [VALUE]
	 *
	 * @return a string of data allocated with malloc().
	 * 	It will be deleted using free().
	 */
	virtual char* writeStatistics();


	void reset();


	/**
	 * Initialize to a custom starting point.
	 */
	void initToGR6N20A();
};

#endif /* ORBITALBRANCHINGMANAGER_HPP_ */
