/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * SBRConnected.cpp
 *
 *  Created on: Aug 24, 2011
 *      Author: derrickstolee
 */

#include "SaturationData.hpp"
#include "SaturationNode.hpp"
#include "SaturationBranchingRule.hpp"
#include "SBRConnected.hpp"
#include "translation.hpp"

/**
 * Constructor
 *
 * Takes an enum for the "sub" branching rule: how to select within the "connected" classes.
 */
SBRConnected::SBRConnected(SubBranchingRule rule)
{
	this->sub = rule;
}

SBRConnected::~SBRConnected()
{

}

/**
 * selectOrbitIndex
 *
 * Perform the implemented branching rule to select an orbit
 * 	index from the given list of orbits.
 * The input orbits should be orbits corresponding to pair-orbits
 * 	of "unknown" edges in the current partial graph.
 *
 * @param data the GraphData object containing all of the graph data.
 * @param node the SaturationNode object containing all of the orbit data.
 *
 */
int SBRConnected::selectOrbitIndex(GraphData* data, SaturationNode* node)
{
	int best_index = 0;
	int best_size = node->orbitSizes[0];
	int best_connectivity = 0;

	int vi, vj;

	indexToPair(node->orbitReps[0], vi, vj);

	int best_degree = data->getDegree(vi, 1) + data->getDegree(vj, 1);

	if ( data->isInConnected(vi) || data->isInConnected(vj) )
	{
		best_connectivity = 1;
	}

	if ( data->isInConnected(vi) && data->isInConnected(vj) )
	{
		best_connectivity = 2;
	}

	for ( int a = 1; a < node->numOrbits; a++ )
	{
		indexToPair(node->orbitReps[a], vi, vj);
		if ( data->isInConnected(vi) && data->isInConnected(vj) )
		{
			if ( best_connectivity < 2 )
			{
				best_index = a;
				best_connectivity = 2;
				best_size = node->orbitSizes[a];
				best_degree = data->getDegree(vi, 1) + data->getDegree(vj, 1);
			}
			else
			{
				/* break ties! */
				switch ( this->sub )
				{
				case BIGGEST:
					if ( node->orbitSizes[a] > best_size )
					{
						best_size = node->orbitSizes[a];
						best_index = a;
					}
					break;

				case MAX_DEGREE:
					int sum = data->getDegree(vi, 1) + data->getDegree(vj, 1);
					if ( sum > best_degree )
					{
						best_degree = sum;
						best_size = node->orbitSizes[a];
						best_index = a;
					}
					break;
				}
			}
		}
		else if ( best_connectivity == 1 && (data->isInConnected(vi) || data->isInConnected(vj)) )
		{
			/* break ties! */
			switch ( this->sub )
			{
			case BIGGEST:
				if ( node->orbitSizes[a] > best_size )
				{
					best_size = node->orbitSizes[a];
					best_index = a;
				}
				break;

			case MAX_DEGREE:
				int sum = data->getDegree(vi, 1) + data->getDegree(vj, 1);
				if ( sum > best_degree )
				{
					best_degree = sum;
					best_index = a;
				}
				break;
			}
		}
		else if ( best_connectivity == 0 &&  (data->isInConnected(vi) || data->isInConnected(vj)))
		{
			best_connectivity = 1;
			best_size = node->orbitSizes[a];
			best_index = a;
			best_degree = data->getDegree(vi, 1) + data->getDegree(vj, 1);
		}
	}

	return best_index;
}
