/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * SaturationData.cpp
 *
 *  Created on: Apr 14, 2011
 *      Author: stolee
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "nausparse.h"
#include "gtools.h"
#include "connectivity.hpp"
#include "translation.hpp"
#include "TreeSet.hpp"
#include "GraphData.hpp"
#include "SaturationData.hpp"
#include "SaturationDataSymmetry.hpp"

#define CHECKANDFREE(STUFF) if ( STUFF != 0 ){ free(STUFF); STUFF = 0; }
#define CHECKANDDELETE(STUFF) if ( STUFF != 0 ){ delete STUFF; STUFF = 0; }
#define POPALL(STACK) while (STACK.size() > 0) { STACK.pop(); }

#define SD_DEBUG_MODE
#define SD_DEBUG_PRIORITY 3

#ifdef SD_DEBUG_MODE
#define SD_DEBUG1(PRIORITY,STRING,ARG) if ( PRIORITY <= SD_DEBUG_PRIORITY) printf(STRING,ARG)
#define SD_DEBUG2(PRIORITY,STRING,ARG1,ARG2) if ( PRIORITY <= SD_DEBUG_PRIORITY) printf(STRING,ARG1,ARG2)
#define SD_DEBUG3(PRIORITY,STRING,ARG1,ARG2,ARG3) if ( PRIORITY <= SD_DEBUG_PRIORITY) printf(STRING,ARG1,ARG2,ARG3)
#else
#define SD_DEBUG1(PRIORITY,STRING,ARG)
#define SD_DEBUG2(PRIORITY,STRING,ARG1,ARG2)
#define SD_DEBUG3(PRIORITY,STRING,ARG1,ARG2,ARG3)
#endif

/**
 * edges_in_set stores the number of edges within a set of order (r-2)
 *
 * This is modified by the methods addEdgeInSet(size, index) and
 * 	removeEdgeInSet(size, index).
 *
 * Returns FALSE if a K_r is created.
 */
//	char* edges_in_set;
void SaturationData::initEdgesInSet()
{
	int nchooses = nChooseK(this->n, this->r - 2);
	this->edges_in_set = (char*) malloc(nchooses);

	for ( int i = 0; i < nchooses; i++ )
	{
		this->edges_in_set[i] = 0;
	}
}

void SaturationData::cleanEdgesInSet()
{
	CHECKANDFREE(this->edges_in_set);
	POPALL(this->edges_in_set_stack);
	POPALL(this->edges_in_set_sizes);
}

int SaturationData::getMaxEdgesInSetSize()
{
	return ((this->r - 2) * (this->r - 3)) / 2;
}

bool SaturationData::addEdgeInSet(int index)
{
	if ( index < 0 || index >= this->nchooses )
	{
		SD_DEBUG1(1, "--addEdgeInSet(%d) out of bounds!\n", index);
		return false;
	}

	this->edges_in_set[index] = this->edges_in_set[index] + 1;
	this->edges_in_set_stack.push(index);

	if ( this->edges_in_set[index] == this->getMaxEdgesInSetSize() )
	{
		/* we have filled this! */
		/* Are there edges which now must be zero? */
		this->edges_complete_to_set[index]->resetIterator();

		bool success = true;
		while ( success && this->edges_complete_to_set[index]->hasNext() )
		{
			int edge = this->edges_complete_to_set[index]->next();

			char e_type = this->getAdjacency(edge);

			if ( e_type == 1 )
			{
				/* we now have a K_r! */
				success = false;
				this->inconsistent = true;
			}
			else if ( e_type == 2 )
			{
				success = success && ((GraphData*) this)->augmentEdge(edge, 0);
				success = this->setCompletionSet(edge, index);
			}
			else if ( e_type == 0 )
			{
				success = this->setCompletionSet(edge, index);
			}
		}

		if ( !success )
		{
			return false;
		}
	}

	return true;
}

bool SaturationData::isSetFull(int index)
{
	return this->edges_in_set[index] == this->getMaxEdgesInSetSize();
}

//	std::stack<int> edges_in_set_stack;
//	std::stack<int> edges_in_set_sizes;
void SaturationData::snapshotEdgesInSet()
{
	this->edges_in_set_sizes.push(this->edges_in_set_stack.size());
}

void SaturationData::rollbackEdgesInSet()
{
	int to_size = this->edges_in_set_sizes.top();
	this->edges_in_set_sizes.pop();

	while ( this->edges_in_set_stack.size() > to_size )
	{
		int index = this->edges_in_set_stack.top();
		this->edges_in_set_stack.pop();

		this->edges_in_set[index] = this->edges_in_set[index] - 1;
	}
}

/**
 * edges_to_set stores the number of edges between a set of order 2...(r-1)
 * 	and another vertex i
 */
//	char* edges_to_set;
//	Set* edges_complete_to_set; /* store a set of edge indices for each set */
//	Set* sets_complete_to_edges; /* store a set of set indices for each edge */
void SaturationData::initEdgesToSet()
{
	int ets_size = this->nchoose2 * this->nchooses;
	this->edges_to_set = (char*) malloc(ets_size);
	bzero(this->edges_to_set, ets_size);

	this->edges_complete_to_set = (Set**) malloc(this->nchooses * sizeof(Set*));
	for ( int i = 0; i < this->nchooses; i++ )
	{
		this->edges_complete_to_set[i] = new TreeSet();
	}

	this->sets_complete_to_edges = (Set**) malloc(this->nchoose2 * sizeof(Set*));
	for ( int i = 0; i < this->nchoose2; i++ )
	{
		this->sets_complete_to_edges[i] = new TreeSet();
	}
}

void SaturationData::cleanEdgesToSet()
{
	CHECKANDFREE(this->edges_to_set);

	if ( this->sets_complete_to_edges != 0 )
	{
		for ( int i = 0; i < this->nchoose2; i++ )
		{
			CHECKANDDELETE(this->sets_complete_to_edges[i])
		}
		free(this->sets_complete_to_edges);
		this->sets_complete_to_edges = 0;
	}

	if ( this->edges_complete_to_set != 0 )
	{
		for ( int i = 0; i < this->nchooses; i++ )
		{
			CHECKANDDELETE(this->edges_complete_to_set[i])
		}

		free(this->edges_complete_to_set);
		this->edges_complete_to_set = 0;
	}

	POPALL(this->edges_to_set_edge_stack);
	POPALL(this->edges_to_set_set_stack);
	POPALL(this->edges_to_set_sizes);
}

int SaturationData::getEdgesToSetPosition(int index, int edge)
{
	return index * this->nchoose2 + edge;
}

int SaturationData::getMaxEdgesToSetSize()
{
	return this->s * 2;
}

bool SaturationData::addEdgeToSet(int index, int edge)
{
	int pos = this->getEdgesToSetPosition(index, edge);

	this->edges_to_set[pos] = this->edges_to_set[pos] + 1;
	this->edges_to_set_edge_stack.push(edge);
	this->edges_to_set_set_stack.push(index);

	if ( this->edges_to_set[pos] == this->getMaxEdgesToSetSize() )
	{
		/* trigger a full set? */
		this->sets_complete_to_edges[edge]->add(index);
		this->edges_complete_to_set[index]->add(edge);

		bool success = true;

		/* can only trigger between these two */
		if ( this->isSetFull(index) )
		{
			/* this edge better be 2! */
			char e_type = this->getAdjacency(edge);

			if ( e_type == 2 )
			{
				/* make a 0-edge */
				success = ((GraphData*) this)->augmentEdge(edge, 0);
				success = this->setCompletionSet(edge, index);
			}
			else if ( e_type == 1 )
			{
				/* we have a K_r! */
				success = false;
				this->inconsistent = true;
			}
			else
			{
				success = this->setCompletionSet(edge, index);
			}
		}

		return success;
	}

	return true;
}

bool SaturationData::doesSetDominate(int index, int edge)
{
	return this->edges_to_set[this->getEdgesToSetPosition(index, edge)] == this->getMaxEdgesToSetSize();
}

//	std::stack<int> edges_to_set_edge_stack;
//	std::stack<int> edges_to_set_set_stack;
//	std::stack<int> edges_to_set_sizes;
void SaturationData::snapshotEdgesToSet()
{
	this->edges_to_set_sizes.push(this->edges_to_set_edge_stack.size());
}

void SaturationData::rollbackEdgesToSet()
{
	int to_size = this->edges_to_set_sizes.top();
	this->edges_to_set_sizes.pop();

	while ( this->edges_to_set_edge_stack.size() > to_size )
	{
		int edge = this->edges_to_set_edge_stack.top();
		this->edges_to_set_edge_stack.pop();

		int index = this->edges_to_set_set_stack.top();
		this->edges_to_set_set_stack.pop();

		int pos = this->getEdgesToSetPosition(index, edge);

		this->edges_to_set[pos] = this->edges_to_set[pos] - 1;

		this->edges_complete_to_set[index]->remove(edge);
		this->sets_complete_to_edges[edge]->remove(index);
	}
}

/**
 * completion_sets stores the index of the (r-2)-set which
 * 	completes the given edge. Value is -1 if there is no completion.
 */
//	int* completion_sets;
void SaturationData::initCompletionSets()
{
	this->completion_sets = (int*) malloc(this->nchoose2 * sizeof(int));

	for ( int i = 0; i < this->nchoose2; i++ )
	{
		this->completion_sets[i] = -1;
	}
}

void SaturationData::cleanCompletionSets()
{
	CHECKANDFREE(this->completion_sets);

	POPALL(this->completion_sets_stack);
	POPALL(this->completion_sets_sizes);
}

int SaturationData::getCompletionSet(int i, int j)
{
	return this->getCompletionSet(indexOf(i, j));
}

int SaturationData::getCompletionSet(int index)
{
	return this->completion_sets[index];
}

bool SaturationData::setCompletionSet(int i, int j, int completionIndex)
{
	return this->setCompletionSet(indexOf(i, j), completionIndex);
}

bool SaturationData::setCompletionSet(int edge, int completionIndex)
{
	if ( this->completion_sets[edge] == -1 )
	{
		this->completion_sets[edge] = completionIndex;

		this->completion_sets_stack.push(edge);

		return true;
	}

	this->inconsistent = true;
	return false;
}

//	std::stack<int> completion_sets_stack;
//	std::stack<int> completion_sets_sizes;
void SaturationData::snapshotCompletionSets()
{
	this->completion_sets_sizes.push(this->completion_sets_stack.size());
}

void SaturationData::rollbackCompletionSets()
{
	int to_size = this->completion_sets_sizes.top();
	this->completion_sets_sizes.pop();

	while ( this->completion_sets_stack.size() > to_size )
	{
		int edge = this->completion_sets_stack.top();
		this->completion_sets_stack.pop();

		this->completion_sets[edge] = -1;
	}
}

void SaturationData::init()
{
	this->inconsistent = false;
	super::init();

	this->initEdgesInSet();
	this->initEdgesToSet();
	this->initCompletionSets();
}

void SaturationData::clean()
{
	super::clean();
	this->cleanEdgesInSet();
	this->cleanEdgesToSet();
	this->cleanCompletionSets();
}

/**
 * Constructor
 */
SaturationData::SaturationData(int r, int N) :
	GraphData(r, N)
{
	this->initEdgesInSet();
	this->initEdgesToSet();
	this->initCompletionSets();
}

/**
 * Destructor
 */
SaturationData::~SaturationData()
{
	this->cleanEdgesInSet();
	this->cleanEdgesToSet();
	this->cleanCompletionSets();
}

/**
 * snapshot()
 *
 * Freeze the current state in a stack.
 */
void SaturationData::snapshot()
{
	super::snapshot();
	this->snapshotEdgesInSet();
	this->snapshotEdgesToSet();
	this->snapshotCompletionSets();
}

/**
 * rollback()
 *
 * Pop the stack and revert the structure to the previous state.
 */
void SaturationData::rollback()
{
	super::rollback();
	this->rollbackEdgesInSet();
	this->rollbackEdgesToSet();
	this->rollbackCompletionSets();
	this->inconsistent = false;
}

/**
 * reset()
 *
 * Set the object back to the initial configuration.
 */
void SaturationData::reset()
{
	this->clean();
	this->init();
}

/**
 * augmentEdge()
 *
 * Augment the graph using the following edge change.
 *
 * return FALSE if the constraints are violated.
 */
bool SaturationData::augmentEdge(int i, int j, char type)
{
	if ( type != 0 && type != 1 )
	{
		return false;
	}

	int edge = indexOf(i, j);

	if ( this->getAdjacency(edge) != 2 )
	{
		if ( this->getAdjacency(edge) == type )
		{
			return true;
		}
		else
		{
			this->inconsistent = true;
			return false;
		}
	}

	this->addConnectedVert(i);
	this->addConnectedVert(j);

	bool success = true;

	/** Graph Stuff! **/
	this->addAdjacency(i, j, type);
	this->addEdgeToLayerGraph(i, j, type);

	if ( type == 1 )
	{
		this->addEdgeToOneGraph(i, j);
	}
	else if ( type == 0 )
	{
		this->addEdgeToZeroGraph(i, j);
	}

	this->addDegree(i, type);
	this->addDegree(j, type);

	/** Sets! **/
	this->removeTwoEdge(edge);

	if ( type == 0 )
	{
		this->addZeroEdge(edge);
	}
	else if ( type == 1 )
	{
		/* Check the edges-in-set */
		int* set = (int*) malloc(this->s * sizeof(int));

		for ( int wwindex = 0; success && wwindex < numSetsWW(this->n, this->s, i, j); wwindex++ )
		{
			getSetWW(this->s, i, j, wwindex, set);

			int index = indexOfSet(this->s, set);

			success = success && this->addEdgeInSet(index);
		}

		/* Check the edges-to-set */
		for ( int wwoindex = 0; success && wwoindex < numSetsWWO(this->n, this->s, i, j); wwoindex++ )
		{
			/* with i, without j */
			getSetWWO(this->s, i, j, wwoindex, set);

			int index = indexOfSet(this->s, set);

			for ( int k = 0; success && k < this->n; k++ )
			{
				bool kinset = false;
				for ( int a = 0; a < this->s; a++ )
				{
					if ( set[a] == k )
					{
						kinset = true;
						break;
					}
				}

				if ( !kinset && k != j )
				{
					success = success && this->addEdgeToSet(index, indexOf(j, k));
				}
			}

			/* with i, without j */
			getSetWWO(this->s, j, i, wwoindex, set);

			index = indexOfSet(this->s, set);

			for ( int k = 0; success && k < this->n; k++ )
			{
				bool kinset = false;
				for ( int a = 0; a < this->s; a++ )
				{
					if ( set[a] == k )
					{
						kinset = true;
						break;
					}
				}

				if ( !kinset && k != i )
				{
					success = success && this->addEdgeToSet(index, indexOf(i, k));
				}
			}
		}

		free(set);
		set = 0;
	}

	return success;
}

/**
 * isUniquelyKrSaturated
 */
bool SaturationData::isUniquelyKrSaturated()
{
	if ( this->inconsistent )
	{
		return false;
	}

	if ( this->two_edges->size() > 0 )
	{
		return false;
	}

	/* check that all 0-edges have a completion */
	//	for ( this->zero_edges->resetIterator(); this->zero_edges->hasNext(); )
	//	{
	//		if ( this->completion_sets[this->zero_edges->next()] < 0 )
	//		{
	//			return false;
	//		}
	//	}
	return true;
}

