/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * SaturationData.hpp
 *
 *  Created on: Apr 14, 2011
 *      Author: stolee
 */

#ifndef SATURATIONDATA_HPP_
#define SATURATIONDATA_HPP_

#include <stack>
#include "nausparse.h"
#include "Set.hpp"
#include "GraphData.hpp"

/**
 * SaturationData is a second attempt at storing data
 * for the Saturation project. It stores all graph information
 * as well as data structures which find complete subgraphs and
 * multiple K_r completions.
 *
 * It behaves in a stack-like fashion, storing a stack of
 * 	operations which can be reversed.
 * However, these operations are stored edge-by-edge.
 *
 * The 'snapshot()' method stores the current data point as
 * 	if it was a stack entry. Any edge operations after this
 * point are stored in the stack.
 *
 * The 'rollback()' method reverses the edge operations
 * 	until reaching the position of the top 'snapshot()' call.
 */
class SaturationData : public GraphData
{
protected:
	typedef GraphData super;

	/**
	 * edges_in_set stores the number of edges within a set of order (r-2)
	 *
	 * This is modified by the methods addEdgeInSet(size, index) and
	 * 	removeEdgeInSet(size, index).
	 *
	 * Returns FALSE if a K_r is created.
	 */
	char* edges_in_set;
	void initEdgesInSet();
	void cleanEdgesInSet();
	int getMaxEdgesInSetSize();
	bool addEdgeInSet(int index);
	void removeEdgeInSet(int index);
	bool isSetFull(int index);
	std::stack<int> edges_in_set_stack;
	std::stack<int> edges_in_set_sizes;
	void snapshotEdgesInSet();
	void rollbackEdgesInSet();

	/**
	 * edges_to_set stores the number of edges between a set of order 2...(r-1)
	 * 	and another vertex i
	 */
	char* edges_to_set;
	Set** edges_complete_to_set; /* store a set of edge indices for each set */
	Set** sets_complete_to_edges; /* store a set of set indices for each edge */
	void initEdgesToSet();
	void cleanEdgesToSet();
	int getEdgesToSetPosition(int index, int edge);
	int getMaxEdgesToSetSize();
	bool addEdgeToSet(int index, int edge);
	bool doesSetDominate(int index, int edge);
	std::stack<int> edges_to_set_edge_stack;
	std::stack<int> edges_to_set_set_stack;
	std::stack<int> edges_to_set_sizes;
	void snapshotEdgesToSet();
	void rollbackEdgesToSet();

	/**
	 * completion_sets stores the index of the (r-2)-set which
	 * 	completes the given edge. Value is -1 if there is no completion.
	 */
	int* completion_sets;
	void initCompletionSets();
	void cleanCompletionSets();
	int getCompletionSet(int i, int j);
	int getCompletionSet(int index);
	bool setCompletionSet(int i, int j, int completionIndex);
	bool setCompletionSet(int edge, int completionIndex);
	std::stack<int> completion_sets_stack;
	std::stack<int> completion_sets_sizes;
	void snapshotCompletionSets();
	void rollbackCompletionSets();

	virtual void init();
	virtual void clean();

public:
	/**
	 * Constructor
	 */
	SaturationData(int r, int N);

	/**
	 * Destructor
	 */
	virtual ~SaturationData();

	/**
	 * snapshot()
	 *
	 * Freeze the current state in a stack.
	 */
	void snapshot();

	/**
	 * rollback()
	 *
	 * Pop the stack and revert the structure to the previous state.
	 */
	void rollback();

	/**
	 * reset()
	 *
	 * Set the object back to the initial configuration.
	 */
	void reset();

	/**
	 * augmentEdge()
	 *
	 * Augment the graph using the following edge change.
	 *
	 * return FALSE if the constraints are violated.
	 */
	bool augmentEdge(int i, int j, char type);


	/**
	 * isUniquelyKrSaturated
	 */
	bool isUniquelyKrSaturated();

};

#endif /* SATURATIONDATA_HPP_ */
